import re
import subprocess


class CommitNotFoundException(Exception):
    pass


def git_remote():
    git_repo = subprocess.run(
        ["git", "remote", "get-url", "origin"], capture_output=True, text=True
    ).stdout.strip()
    return extract_repository_name(git_repo)


def extract_repository_name(repository_url):
    if not repository_url:
        return

    return re.search(r"([^/:]*/[^/]*)\.git", repository_url).group(1)


def check_if_commit_exists(commit):
    branches_containing_commit = subprocess.run(
        ["git", "branch", "-r", "--contains", f"{commit}"], capture_output=True, text=True
    )

    if branches_containing_commit.stderr:
        raise CommitNotFoundException()



import re
from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import Optional
from typing import Union

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_FILE
from dbt_platform_helper.providers.validation import ValidationException


class IncompatibleMajorVersionException(ValidationException):
    def __init__(self, app_version: str, check_version: str):
        super().__init__()
        self.app_version = app_version
        self.check_version = check_version


class IncompatibleMinorVersionException(ValidationException):
    def __init__(self, app_version: str, check_version: str):
        super().__init__()
        self.app_version = app_version
        self.check_version = check_version


class SemanticVersion:
    def __init__(self, major, minor, patch):
        self.major = major
        self.minor = minor
        self.patch = patch

    def __str__(self) -> str:
        if self.major is None:
            return "unknown"
        return ".".join([str(s) for s in [self.major, self.minor, self.patch]])

    def __lt__(self, other) -> bool:
        return (self.major, self.minor, self.patch) < (other.major, other.minor, other.patch)

    def __eq__(self, other) -> bool:
        return (self.major, self.minor, self.patch) == (other.major, other.minor, other.patch)

    def validate_compatibility_with(self, other):
        if (self.major == 0 and other.major == 0) and (
            self.minor != other.minor or self.patch != other.patch
        ):
            raise IncompatibleMajorVersionException(str(self), str(other))

        if self.major != other.major:
            raise IncompatibleMajorVersionException(str(self), str(other))

        if self.minor != other.minor:
            raise IncompatibleMinorVersionException(str(self), str(other))

    @staticmethod
    def from_string(version_string: Union[str, None]):
        if version_string is None:
            return None

        version_plain = version_string.replace("v", "")
        version_segments = re.split(r"[.\-]", version_plain)

        if len(version_segments) != 3:
            return None

        output_version = [0, 0, 0]
        for index, segment in enumerate(version_segments):
            try:
                output_version[index] = int(segment)
            except ValueError:
                output_version[index] = -1

        return SemanticVersion(output_version[0], output_version[1], output_version[2])


class VersionStatus:
    def __init__(
        self, local_version: SemanticVersion = None, latest_release: SemanticVersion = None
    ):
        self.local = local_version
        self.latest = latest_release

    def is_outdated(self):
        return self.local != self.latest

    def warn(self):
        pass


@dataclass
class PlatformHelperVersionStatus(VersionStatus):
    local: Optional[SemanticVersion] = None
    latest: Optional[SemanticVersion] = None
    deprecated_version_file: Optional[SemanticVersion] = None
    platform_config_default: Optional[SemanticVersion] = None
    pipeline_overrides: Optional[Dict[str, str]] = field(default_factory=dict)

    def warn(self) -> dict:
        if self.platform_config_default and not self.deprecated_version_file:
            return {}

        warnings = []
        errors = []

        missing_default_version_message = f"Create a section in the root of '{PLATFORM_CONFIG_FILE}':\n\ndefault_versions:\n  platform-helper: "
        deprecation_message = (
            f"Please delete '{PLATFORM_HELPER_VERSION_FILE}' as it is now deprecated."
        )

        if self.platform_config_default and self.deprecated_version_file:
            warnings.append(deprecation_message)

        if not self.platform_config_default and self.deprecated_version_file:
            warnings.append(deprecation_message)
            warnings.append(f"{missing_default_version_message}{self.deprecated_version_file}\n")

        if not self.platform_config_default and not self.deprecated_version_file:
            message = f"Cannot get dbt-platform-helper version from '{PLATFORM_CONFIG_FILE}'.\n"
            message += f"{missing_default_version_message}{self.local}\n"
            errors.append(message)

        return {
            "warnings": warnings,
            "errors": errors,
        }



from typing import Callable

import boto3

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.opensearch import OpensearchProvider
from dbt_platform_helper.providers.redis import RedisProvider


class ConfigValidatorError(PlatformException):
    pass


class ConfigValidator:

    def __init__(
        self, validations: Callable[[dict], None] = None, io: ClickIOProvider = ClickIOProvider()
    ):
        self.validations = validations or [
            self.validate_supported_redis_versions,
            self.validate_supported_opensearch_versions,
            self.validate_environment_pipelines,
            self.validate_environment_pipelines_triggers,
            self.validate_database_copy_section,
            self.validate_database_migration_input_sources,
        ]
        self.io = io

    def run_validations(self, config: dict):
        for validation in self.validations:
            validation(config)

    def _validate_extension_supported_versions(
        self, config, extension_type, version_key, get_supported_versions
    ):
        extensions = config.get("extensions", {})
        if not extensions:
            return

        extensions_for_type = [
            extension
            for extension in config.get("extensions", {}).values()
            if extension.get("type") == extension_type
        ]

        supported_extension_versions = get_supported_versions()
        extensions_with_invalid_version = []

        for extension in extensions_for_type:

            environments = extension.get("environments", {})

            if not isinstance(environments, dict):
                self.io.error(
                    f"{extension_type} extension definition is invalid type, expected dictionary",
                )
                continue
            for environment, env_config in environments.items():

                # An extension version doesn't need to be specified for all environments, provided one is specified under "*".
                # So check if the version is set before checking if it's supported
                extension_version = env_config.get(version_key)
                if extension_version and extension_version not in supported_extension_versions:
                    extensions_with_invalid_version.append(
                        {"environment": environment, "version": extension_version}
                    )

        for version_failure in extensions_with_invalid_version:
            self.io.error(
                f"{extension_type} version for environment {version_failure['environment']} is not in the list of supported {extension_type} versions: {supported_extension_versions}. Provided Version: {version_failure['version']}",
            )

    def validate_supported_redis_versions(self, config):
        return self._validate_extension_supported_versions(
            config=config,
            extension_type="redis",
            version_key="engine",
            get_supported_versions=RedisProvider(
                boto3.client("elasticache")
            ).get_supported_redis_versions,
        )

    def validate_supported_opensearch_versions(self, config):
        return self._validate_extension_supported_versions(
            config=config,
            extension_type="opensearch",
            version_key="engine",
            get_supported_versions=OpensearchProvider(
                boto3.client("opensearch")
            ).get_supported_opensearch_versions,
        )

    def validate_environment_pipelines(self, config):
        bad_pipelines = {}
        for pipeline_name, pipeline in config.get("environment_pipelines", {}).items():
            bad_envs = []
            pipeline_account = pipeline.get("account", None)
            if pipeline_account:
                for env in pipeline.get("environments", {}).keys():
                    env_account = (
                        config.get("environments", {})
                        .get(env, {})
                        .get("accounts", {})
                        .get("deploy", {})
                        .get("name")
                    )
                    if not env_account == pipeline_account:
                        bad_envs.append(env)
            if bad_envs:
                bad_pipelines[pipeline_name] = {"account": pipeline_account, "bad_envs": bad_envs}
        if bad_pipelines:
            message = "The following pipelines are misconfigured:"
            for pipeline, detail in bad_pipelines.items():
                envs = detail["bad_envs"]
                acc = detail["account"]
                message += f"  '{pipeline}' - these environments are not in the '{acc}' account: {', '.join(envs)}\n"
            raise ConfigValidatorError(message)

    def validate_environment_pipelines_triggers(self, config):
        errors = []
        pipelines_with_triggers = {
            pipeline_name: pipeline
            for pipeline_name, pipeline in config.get("environment_pipelines", {}).items()
            if "pipeline_to_trigger" in pipeline
        }

        for pipeline_name, pipeline in pipelines_with_triggers.items():
            pipeline_to_trigger = pipeline["pipeline_to_trigger"]
            if pipeline_to_trigger not in config.get("environment_pipelines", {}):
                message = f"  '{pipeline_name}' - '{pipeline_to_trigger}' is not a valid target pipeline to trigger"

                errors.append(message)
                continue

            if pipeline_to_trigger == pipeline_name:
                message = f"  '{pipeline_name}' - pipelines cannot trigger themselves"
                errors.append(message)

        if errors:
            error_message = "The following pipelines are misconfigured: \n"
            raise ConfigValidatorError(error_message + "\n  ".join(errors))

    def validate_database_copy_section(self, config):
        extensions = config.get("extensions", {})
        if not extensions:
            return

        postgres_extensions = {
            key: ext for key, ext in extensions.items() if ext.get("type", None) == "postgres"
        }

        if not postgres_extensions:
            return

        errors = []

        for extension_name, extension in postgres_extensions.items():
            database_copy_sections = extension.get("database_copy", [])

            if not database_copy_sections:
                return

            all_environments = [
                env for env in config.get("environments", {}).keys() if not env == "*"
            ]
            all_envs_string = ", ".join(all_environments)

            for section in database_copy_sections:
                from_env = section["from"]
                to_env = section["to"]

                from_account = (
                    config.get("environments", {})
                    .get(from_env, {})
                    .get("accounts", {})
                    .get("deploy", {})
                    .get("id")
                )
                to_account = (
                    config.get("environments", {})
                    .get(to_env, {})
                    .get("accounts", {})
                    .get("deploy", {})
                    .get("id")
                )

                if from_env == to_env:
                    errors.append(
                        f"database_copy 'to' and 'from' cannot be the same environment in extension '{extension_name}'."
                    )

                if "prod" in to_env:
                    errors.append(
                        f"Copying to a prod environment is not supported: database_copy 'to' cannot be '{to_env}' in extension '{extension_name}'."
                    )

                if from_env not in all_environments:
                    errors.append(
                        f"database_copy 'from' parameter must be a valid environment ({all_envs_string}) but was '{from_env}' in extension '{extension_name}'."
                    )

                if to_env not in all_environments:
                    errors.append(
                        f"database_copy 'to' parameter must be a valid environment ({all_envs_string}) but was '{to_env}' in extension '{extension_name}'."
                    )

                if from_account != to_account:
                    if "from_account" not in section:
                        errors.append(
                            f"Environments '{from_env}' and '{to_env}' are in different AWS accounts. The 'from_account' parameter must be present."
                        )
                    elif section["from_account"] != from_account:
                        errors.append(
                            f"Incorrect value for 'from_account' for environment '{from_env}'"
                        )

                    if "to_account" not in section:
                        errors.append(
                            f"Environments '{from_env}' and '{to_env}' are in different AWS accounts. The 'to_account' parameter must be present."
                        )
                    elif section["to_account"] != to_account:
                        errors.append(
                            f"Incorrect value for 'to_account' for environment '{to_env}'"
                        )

        if errors:
            raise ConfigValidatorError("\n".join(errors))

    def validate_database_migration_input_sources(self, config: dict):
        extensions = config.get("extensions", {})
        if not extensions:
            return

        s3_extensions = {
            key: ext for key, ext in extensions.items() if ext.get("type", None) == "s3"
        }

        if not s3_extensions:
            return

        errors = []

        for extension_name, extension in s3_extensions.items():
            for env, env_config in extension.get("environments", {}).items():
                if "data_migration" not in env_config:
                    continue
                data_migration = env_config.get("data_migration", {})
                if "import" in data_migration and "import_sources" in data_migration:
                    errors.append(
                        f"Error in '{extension_name}.environments.{env}.data_migration': only the 'import_sources' property is required - 'import' is deprecated."
                    )
                if "import" not in data_migration and "import_sources" not in data_migration:
                    errors.append(
                        f"'import_sources' property in '{extension_name}.environments.{env}.data_migration' is missing."
                    )
        if errors:
            raise ConfigValidatorError("\n".join(errors))



import os
import re
import subprocess
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version
from pathlib import Path

from dbt_platform_helper.constants import DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION
from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_FILE
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.semantic_version import (
    IncompatibleMajorVersionException,
)
from dbt_platform_helper.providers.semantic_version import (
    IncompatibleMinorVersionException,
)
from dbt_platform_helper.providers.semantic_version import PlatformHelperVersionStatus
from dbt_platform_helper.providers.semantic_version import SemanticVersion
from dbt_platform_helper.providers.semantic_version import VersionStatus
from dbt_platform_helper.providers.validation import ValidationException
from dbt_platform_helper.providers.version import GithubVersionProvider
from dbt_platform_helper.providers.version import PyPiVersionProvider
from dbt_platform_helper.providers.yaml_file import FileProviderException
from dbt_platform_helper.providers.yaml_file import YamlFileProvider


class PlatformHelperVersionNotFoundException(PlatformException):
    def __init__(self):
        super().__init__(f"""Platform helper version could not be resolved.""")


class RequiredVersion:
    def __init__(self, io=None):
        self.io = io or ClickIOProvider()

    def get_required_platform_helper_version(
        self, pipeline: str = None, versions: PlatformHelperVersionStatus = None
    ) -> str:
        if not versions:
            versions = get_platform_helper_versions()
        pipeline_version = versions.pipeline_overrides.get(pipeline)
        version_precedence = [
            pipeline_version,
            versions.platform_config_default,
            versions.deprecated_version_file,
        ]
        non_null_version_precedence = [
            f"{v}" if isinstance(v, SemanticVersion) else v for v in version_precedence if v
        ]

        out = non_null_version_precedence[0] if non_null_version_precedence else None

        if not out:
            raise PlatformHelperVersionNotFoundException

        return out

    def get_required_version(self, pipeline=None):
        version = self.get_required_platform_helper_version(pipeline)
        self.io.info(version)
        return version

    # Used in the generate command
    def check_platform_helper_version_mismatch(self):
        if not running_as_installed_package():
            return

        versions = get_platform_helper_versions()
        platform_helper_file_version = SemanticVersion.from_string(
            self.get_required_platform_helper_version(versions=versions)
        )

        if not versions.local == platform_helper_file_version:
            message = (
                f"WARNING: You are running platform-helper v{versions.local} against "
                f"v{platform_helper_file_version} specified by {PLATFORM_HELPER_VERSION_FILE}."
            )
            self.io.warn(message)


# Resolves all the versions from pypi, config and locally installed version
# echos warnings if anything is incompatible
def get_platform_helper_versions(
    include_project_versions=True, yaml_provider=YamlFileProvider
) -> PlatformHelperVersionStatus:
    try:
        locally_installed_version = SemanticVersion.from_string(version("dbt-platform-helper"))
    except PackageNotFoundError:
        locally_installed_version = None

    latest_release = PyPiVersionProvider.get_latest_version("dbt-platform-helper")

    if not include_project_versions:
        return PlatformHelperVersionStatus(
            local=locally_installed_version,
            latest=latest_release,
        )

    deprecated_version_file = Path(PLATFORM_HELPER_VERSION_FILE)
    try:
        loaded_version = yaml_provider.load(deprecated_version_file)
        version_from_file = SemanticVersion.from_string(loaded_version)
    except FileProviderException:
        version_from_file = None

    platform_config_default, pipeline_overrides = None, {}

    config = ConfigProvider()
    platform_config = config.load_unvalidated_config_file()

    if platform_config:
        platform_config_default = SemanticVersion.from_string(
            platform_config.get("default_versions", {}).get("platform-helper")
        )

        pipeline_overrides = {
            name: pipeline.get("versions", {}).get("platform-helper")
            for name, pipeline in platform_config.get("environment_pipelines", {}).items()
            if pipeline.get("versions", {}).get("platform-helper")
        }

    out = PlatformHelperVersionStatus(
        local=locally_installed_version,
        latest=latest_release,
        deprecated_version_file=version_from_file,
        platform_config_default=platform_config_default,
        pipeline_overrides=pipeline_overrides,
    )

    _process_version_file_warnings(out)

    return out


# Validates the returned PlatformHelperVersionStatus and echos useful warnings
# Could return ValidationMessages (warnings and errors) which are output elsewhere
def _process_version_file_warnings(versions: PlatformHelperVersionStatus, io=ClickIOProvider()):
    messages = versions.warn()

    if messages.get("errors"):
        io.error("\n".join(messages["errors"]))

    if messages.get("warnings"):
        io.warn("\n".join(messages["warnings"]))


# TODO called at the beginning of every command.  This is platform-version base functionality
def check_platform_helper_version_needs_update(io=ClickIOProvider()):
    if not running_as_installed_package() or "PLATFORM_TOOLS_SKIP_VERSION_CHECK" in os.environ:
        return
    versions = get_platform_helper_versions(include_project_versions=False)
    local_version = versions.local
    latest_release = versions.latest
    message = (
        f"You are running platform-helper v{local_version}, upgrade to "
        f"v{latest_release} by running run `pip install "
        "--upgrade dbt-platform-helper`."
    )
    try:
        local_version.validate_compatibility_with(latest_release)
    except IncompatibleMajorVersionException:
        io.error(message)
    except IncompatibleMinorVersionException:
        io.warn(message)


# TODO can stay as utility for now
def running_as_installed_package():
    return "site-packages" in __file__


def get_required_terraform_platform_modules_version(
    cli_terraform_platform_modules_version, platform_config_terraform_modules_default_version
):
    version_preference_order = [
        cli_terraform_platform_modules_version,
        platform_config_terraform_modules_default_version,
        DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION,
    ]
    return [version for version in version_preference_order if version][0]


#########################################
# Only used in Config domain
# TODO to be relocated along with tests
#########################################


# Getting version from the "Generated by" comment in a file that was generated from a template
# TODO where does this belong?  It sort of belongs to our platform-helper templating
def get_template_generated_with_version(template_file_path: str) -> SemanticVersion:
    try:
        template_contents = Path(template_file_path).read_text()
        template_version = re.match(
            r"# Generated by platform-helper ([v.\-0-9]+)", template_contents
        ).group(1)
        return SemanticVersion.from_string(template_version)
    except (IndexError, AttributeError):
        raise ValidationException(f"Template {template_file_path} has no version information")


def validate_template_version(app_version: SemanticVersion, template_file_path: str):
    app_version.validate_compatibility_with(get_template_generated_with_version(template_file_path))


# Local version and latest release of tool.
# Used only in config command.
# TODO Move to config domain
def get_copilot_versions() -> VersionStatus:
    copilot_version = None

    try:
        response = subprocess.run("copilot --version", capture_output=True, shell=True)
        [copilot_version] = re.findall(r"[0-9.]+", response.stdout.decode("utf8"))
    except ValueError:
        pass

    return VersionStatus(
        SemanticVersion.from_string(copilot_version),
        GithubVersionProvider.get_latest_version("aws/copilot-cli"),
    )


# Local version and latest release of tool.
# Used only in config command.
# TODO Move to config domain
def get_aws_versions() -> VersionStatus:
    aws_version = None
    try:
        response = subprocess.run("aws --version", capture_output=True, shell=True)
        matched = re.match(r"aws-cli/([0-9.]+)", response.stdout.decode("utf8"))
        aws_version = SemanticVersion.from_string(matched.group(1))
    except ValueError:
        pass

    return VersionStatus(aws_version, GithubVersionProvider.get_latest_version("aws/aws-cli", True))



from dbt_platform_helper.providers.cache import CacheProvider


class OpensearchProvider:

    def __init__(self, opensearch_client):
        self.opensearch_client = opensearch_client

    def get_supported_opensearch_versions(self) -> list[str]:

        cache_provider = self.__get_cache_provider()

        if cache_provider.cache_refresh_required("opensearch"):

            response = self.opensearch_client.list_versions()
            all_versions = response["Versions"]

            opensearch_versions = [
                version for version in all_versions if not version.startswith("Elasticsearch_")
            ]
            supported_versions = [
                version.removeprefix("OpenSearch_") for version in opensearch_versions
            ]

            cache_provider.update_cache("opensearch", supported_versions)

            return supported_versions

        else:
            return cache_provider.read_supported_versions_from_cache("opensearch")

    # TODO - cache provider instantiated here rather than via dependancy injection since it will likely only be used in the get_supported_opensearch_versions method.
    # If another method is added which needs a CacheProvider, it should be injected into the constructor instead.
    @staticmethod
    def __get_cache_provider():
        return CacheProvider()



from abc import ABC

import requests

from dbt_platform_helper.providers.semantic_version import SemanticVersion


class VersionProvider(ABC):
    pass


# TODO add timeouts and exception handling for requests
# TODO Alternatively use the gitpython package?
class GithubVersionProvider(VersionProvider):
    @staticmethod
    def get_latest_version(repo_name: str, tags: bool = False) -> SemanticVersion:
        if tags:
            tags_list = requests.get(f"https://api.github.com/repos/{repo_name}/tags").json()
            versions = [SemanticVersion.from_string(v["name"]) for v in tags_list]
            versions.sort(reverse=True)
            return versions[0]

        package_info = requests.get(
            f"https://api.github.com/repos/{repo_name}/releases/latest"
        ).json()
        return SemanticVersion.from_string(package_info["tag_name"])


class PyPiVersionProvider(VersionProvider):
    @staticmethod
    def get_latest_version(project_name: str) -> SemanticVersion:
        package_info = requests.get(f"https://pypi.org/pypi/{project_name}/json").json()
        released_versions = package_info["releases"].keys()
        parsed_released_versions = [SemanticVersion.from_string(v) for v in released_versions]
        parsed_released_versions.sort(reverse=True)
        return parsed_released_versions[0]



from boto3 import Session

from dbt_platform_helper.utils.aws import get_aws_session_or_abort


class ECRProvider:
    def __init__(self, session: Session = None):
        self.session = session
        self.client = None

    def _get_client(self):
        if not self.session:
            self.session = get_aws_session_or_abort()
        return self.session.client("ecr")

    def get_ecr_repo_names(self) -> list[str]:
        out = []
        for page in self._get_client().get_paginator("describe_repositories").paginate():
            out.extend([repo["repositoryName"] for repo in page.get("repositories", {})])
        return out



# Todo: Move to Config provider
PLATFORM_CONFIG_FILE = "platform-config.yml"
# Todo: Can we get rid of this yet?
PLATFORM_HELPER_VERSION_FILE = ".platform-helper-version"
# Todo: Move to ???
DEFAULT_TERRAFORM_PLATFORM_MODULES_VERSION = "7"
SUPPORTED_TERRAFORM_VERSION = "~> 1.8"
SUPPORTED_AWS_PROVIDER_VERSION = "~> 5"

# Keys
CODEBASE_PIPELINES_KEY = "codebase_pipelines"
ENVIRONMENTS_KEY = "environments"
ENVIRONMENT_PIPELINES_KEY = "environment_pipelines"

# Conduit
CONDUIT_ADDON_TYPES = [
    "opensearch",
    "postgres",
    "redis",
]
CONDUIT_DOCKER_IMAGE_LOCATION = "public.ecr.aws/uktrade/tunnel"
HYPHENATED_APPLICATION_NAME = "hyphenated-application-name"
ALPHANUMERIC_ENVIRONMENT_NAME = "alphanumericenvironmentname123"
ALPHANUMERIC_SERVICE_NAME = "alphanumericservicename123"
COPILOT_IDENTIFIER = "c0PIlotiD3ntIF3r"
CLUSTER_NAME_SUFFIX = f"Cluster-{COPILOT_IDENTIFIER}"
SERVICE_NAME_SUFFIX = f"Service-{COPILOT_IDENTIFIER}"
REFRESH_TOKEN_MESSAGE = (
    "To refresh this SSO session run `aws sso login` with the corresponding profile"
)



import os
import webbrowser
from pathlib import Path
from typing import Dict

import boto3
import botocore
import click
from prettytable import PrettyTable

from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.semantic_version import (
    IncompatibleMajorVersionException,
)
from dbt_platform_helper.providers.validation import ValidationException
from dbt_platform_helper.utils import versioning
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.versioning import get_platform_helper_versions

yes = "\033[92m✔\033[0m"
no = "\033[91m✖\033[0m"
maybe = "\033[93m?\033[0m"

RECOMMENDATIONS = {
    "dbt-platform-helper-upgrade": (
        "Upgrade dbt-platform-helper to version {version} `pip install "
        "--upgrade dbt-platform-helper=={version}`."
    ),
    "dbt-platform-helper-upgrade-note": (
        "Post upgrade, run `platform-helper copilot make-addons` to " "update your addon templates."
    ),
    "generic-tool-upgrade": "Upgrade {tool} to version {version}.",
}

SSO_START_URL = "https://uktrade.awsapps.com/start"

AWS_CONFIG = """
#
# uktrade
#

[sso-session uktrade]
sso_start_url = https://uktrade.awsapps.com/start#/
sso_region = eu-west-2
sso_registration_scopes = sso:account:access

[default]
sso_session = uktrade
region = eu-west-2
output = json

"""


@click.group(cls=ClickDocOptGroup)
def config():
    """Perform actions on configuration files."""


@config.command()
def validate():
    """Validate deployment or application configuration."""
    ran_checks = False
    if Path("copilot").exists():
        click.secho("\nDetected a deployment repository", fg="blue")
        deployment()
        ran_checks = True

    if not ran_checks:
        click.secho("Could not find a deployment repository, no checks to run.", fg="red")
        exit(1)


def deployment():
    click.secho()

    compatible = True
    platform_helper_versions = get_platform_helper_versions()
    copilot_versions = versioning.get_copilot_versions()
    aws_versions = versioning.get_aws_versions()
    _check_tool_versions(platform_helper_versions, copilot_versions, aws_versions)
    click.secho("Checking addons templates versions...", fg="blue")

    local_version = platform_helper_versions.local
    latest_release = platform_helper_versions.latest
    addons_templates_table = PrettyTable()
    addons_templates_table.field_names = [
        "Addons Template File",
        "Generated with",
        "Compatible with local?",
        "Compatible with latest?",
    ]
    addons_templates_table.align["Addons Template File"] = "l"

    addons_templates = list(Path("./copilot").glob("**/addons/*"))
    # Sort by template file path
    addons_templates.sort(key=lambda e: str(e))
    # Bring environment addons to the top
    addons_templates.sort(key=lambda e: "environments/" not in str(e))

    recommendations = {}

    ConfigProvider().config_file_check()

    for template_file in addons_templates:
        generated_with_version = maybe
        local_compatible_symbol = yes
        latest_compatible_symbol = yes

        try:
            generated_with_version = versioning.get_template_generated_with_version(
                str(template_file.resolve())
            )
            versioning.validate_template_version(local_version, str(template_file.resolve()))
        except IncompatibleMajorVersionException:
            local_compatible_symbol = no
            compatible = False
            recommendations["dbt-platform-helper-upgrade"] = RECOMMENDATIONS[
                "dbt-platform-helper-upgrade"
            ].format(version=latest_release)
            recommendations["dbt-platform-helper-upgrade-note"] = RECOMMENDATIONS[
                "dbt-platform-helper-upgrade-note"
            ]
        except ValidationException:
            local_compatible_symbol = maybe
            compatible = False
            recommendations["dbt-platform-helper-upgrade"] = RECOMMENDATIONS[
                "dbt-platform-helper-upgrade"
            ].format(version=latest_release)
            recommendations["dbt-platform-helper-upgrade-note"] = RECOMMENDATIONS[
                "dbt-platform-helper-upgrade-note"
            ]

        try:
            generated_with_version = versioning.get_template_generated_with_version(
                str(template_file.resolve())
            )
            versioning.validate_template_version(latest_release, str(template_file.resolve()))
        except IncompatibleMajorVersionException:
            latest_compatible_symbol = no
            compatible = False
        except ValidationException:
            latest_compatible_symbol = maybe
            compatible = False

        addons_templates_table.add_row(
            [
                template_file.relative_to("."),
                (maybe if latest_compatible_symbol is maybe else str(generated_with_version)),
                local_compatible_symbol,
                latest_compatible_symbol,
            ]
        )

    click.secho(addons_templates_table)
    render_recommendations(recommendations)

    exit(0 if compatible else 1)


def _check_tool_versions(platform_helper_versions, copilot_versions, aws_versions):
    click.secho("Checking tooling versions...", fg="blue")
    recommendations = {}

    local_copilot_version = copilot_versions.local
    copilot_latest_release = copilot_versions.latest
    if local_copilot_version is None:
        recommendations["install-copilot"] = (
            "Install AWS Copilot https://aws.github.io/copilot-cli/"
        )

    if aws_versions.local is None:
        recommendations["install-aws"] = "Install AWS CLI https://aws.amazon.com/cli/"

    tool_versions_table = PrettyTable()
    tool_versions_table.field_names = [
        "Tool",
        "Local version",
        "Released version",
        "Running latest?",
    ]
    tool_versions_table.align["Tool"] = "l"

    tool_versions_table.add_row(
        [
            "aws",
            str(aws_versions.local),
            str(aws_versions.latest),
            no if aws_versions.is_outdated() else yes,
        ]
    )
    tool_versions_table.add_row(
        [
            "copilot",
            str(copilot_versions.local),
            str(copilot_versions.latest),
            no if copilot_versions.is_outdated() else yes,
        ]
    )
    tool_versions_table.add_row(
        [
            "dbt-platform-helper",
            str(platform_helper_versions.local),
            str(platform_helper_versions.latest),
            no if platform_helper_versions.is_outdated() else yes,
        ]
    )

    click.secho(tool_versions_table)

    if aws_versions.is_outdated() and "install-aws" not in recommendations:
        recommendations["aws-upgrade"] = RECOMMENDATIONS["generic-tool-upgrade"].format(
            tool="AWS CLI",
            version=str(aws_versions.latest),
        )

    if copilot_versions.is_outdated() and "install-copilot" not in recommendations:
        recommendations["copilot-upgrade"] = RECOMMENDATIONS["generic-tool-upgrade"].format(
            tool="AWS Copilot",
            version=str(copilot_latest_release),
        )

    if platform_helper_versions.is_outdated():
        recommendations["dbt-platform-helper-upgrade"] = RECOMMENDATIONS[
            "dbt-platform-helper-upgrade"
        ].format(version=str(platform_helper_versions.latest))
        recommendations["dbt-platform-helper-upgrade-note"] = RECOMMENDATIONS[
            "dbt-platform-helper-upgrade-note"
        ]

    render_recommendations(recommendations)


def render_recommendations(recommendations: Dict[str, str]):
    if recommendations:
        click.secho("\nRecommendations:\n", bold=True)

        for name, recommendation in recommendations.items():
            if name.endswith("-note"):
                continue
            click.secho(f"  - {recommendation}")
            if recommendations.get(f"{name}-note", False):
                click.secho(f"    {recommendations.get(f'{name}-note')}")

        click.secho()


@config.command()
@click.option("--file-path", "-fp", default="~/.aws/config")
def aws(file_path):
    """
    Writes a local config file containing all the AWS profiles to which the
    logged in user has access.

    If no `--file-path` is specified, defaults to `~/.aws/config`.
    """
    sso_oidc_client = boto3.client("sso-oidc", region_name="eu-west-2")
    sso_client = boto3.client("sso", region_name="eu-west-2")
    oidc_app = create_oidc_application(sso_oidc_client)
    verification_url, device_code = get_device_code(sso_oidc_client, oidc_app, SSO_START_URL)

    if click.confirm(
        "You are about to be redirected to a verification page. You will need to complete sign-in before returning to the command line. Do you want to continue?",
        abort=True,
    ):
        webbrowser.open(verification_url)

    if click.confirm("Have you completed the sign-in process in your browser?", abort=True):
        access_token = get_access_token(device_code, sso_oidc_client, oidc_app)

    aws_config_path = os.path.expanduser(file_path)

    if click.confirm(
        f"This command is destructive and will overwrite file contents at {file_path}. Are you sure you want to continue?",
        abort=True,
    ):
        with open(aws_config_path, "w") as config_file:
            config_file.write(AWS_CONFIG)

            for account in retrieve_aws_accounts(sso_client, access_token):
                config_file.write(f"[profile {account['accountName']}]\n")
                config_file.write("sso_session = uktrade\n")
                config_file.write(f"sso_account_id = {account['accountId']}\n")
                config_file.write("sso_role_name = AdministratorAccess\n")
                config_file.write("region = eu-west-2\n")
                config_file.write("output = json\n")
                config_file.write("\n")


def create_oidc_application(sso_oidc_client):
    print("Creating temporary AWS SSO OIDC application")
    client = sso_oidc_client.register_client(
        clientName="platform-helper",
        clientType="public",
    )
    client_id = client.get("clientId")
    client_secret = client.get("clientSecret")

    return client_id, client_secret


def get_device_code(sso_oidc_client, oidc_application, start_url):
    print("Initiating device code flow")
    authz = sso_oidc_client.start_device_authorization(
        clientId=oidc_application[0],
        clientSecret=oidc_application[1],
        startUrl=start_url,
    )
    url = authz.get("verificationUriComplete")
    deviceCode = authz.get("deviceCode")

    return url, deviceCode


def retrieve_aws_accounts(sso_client, aws_sso_token):
    aws_accounts_response = sso_client.list_accounts(
        accessToken=aws_sso_token,
        maxResults=100,
    )
    if len(aws_accounts_response.get("accountList", [])) == 0:
        raise RuntimeError("Unable to retrieve AWS SSO account list\n")
    return aws_accounts_response.get("accountList")


def get_access_token(device_code, sso_oidc_client, oidc_app):
    try:
        token_response = sso_oidc_client.create_token(
            clientId=oidc_app[0],
            clientSecret=oidc_app[1],
            grantType="urn:ietf:params:oauth:grant-type:device_code",
            deviceCode=device_code,
        )

        return token_response.get("accessToken")

    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] != "AuthorizationPendingException":
            raise e



import boto3

from dbt_platform_helper.platform_exception import PlatformException

# TODO - a good candidate for a dataclass when this is refactored into a class.
# Below methods should also really be refactored to not be so tightly coupled with eachother.


def get_load_balancer_for_application(session: boto3.Session, app: str, env: str) -> str:
    lb_client = session.client("elbv2")

    describe_response = lb_client.describe_load_balancers()
    load_balancers = [lb["LoadBalancerArn"] for lb in describe_response["LoadBalancers"]]

    load_balancers = lb_client.describe_tags(ResourceArns=load_balancers)["TagDescriptions"]

    load_balancer_arn = None
    for lb in load_balancers:
        tags = {t["Key"]: t["Value"] for t in lb["Tags"]}
        if tags.get("copilot-application") == app and tags.get("copilot-environment") == env:
            load_balancer_arn = lb["ResourceArn"]

    if not load_balancer_arn:
        raise LoadBalancerNotFoundException(
            f"No load balancer found for {app} in the {env} environment"
        )

    return load_balancer_arn


def get_https_listener_for_application(session: boto3.Session, app: str, env: str) -> str:
    load_balancer_arn = get_load_balancer_for_application(session, app, env)
    lb_client = session.client("elbv2")
    listeners = lb_client.describe_listeners(LoadBalancerArn=load_balancer_arn)["Listeners"]

    listener_arn = None

    try:
        listener_arn = next(l["ListenerArn"] for l in listeners if l["Protocol"] == "HTTPS")
    except StopIteration:
        pass

    if not listener_arn:
        raise ListenerNotFoundException(f"No HTTPS listener for {app} in the {env} environment")

    return listener_arn


def get_https_certificate_for_application(session: boto3.Session, app: str, env: str) -> str:

    listener_arn = get_https_listener_for_application(session, app, env)
    cert_client = session.client("elbv2")
    certificates = cert_client.describe_listener_certificates(ListenerArn=listener_arn)[
        "Certificates"
    ]

    try:
        certificate_arn = next(c["CertificateArn"] for c in certificates if c["IsDefault"])
    except StopIteration:
        raise CertificateNotFoundException(env)

    return certificate_arn


class LoadBalancerException(PlatformException):
    pass


class LoadBalancerNotFoundException(LoadBalancerException):
    pass


class ListenerNotFoundException(LoadBalancerException):
    pass


class ListenerRuleNotFoundException(LoadBalancerException):
    pass


class CertificateNotFoundException(PlatformException):
    def __init__(self, environment_name: str):
        super().__init__(
            f"""No certificate found with domain name matching environment {environment_name}."."""
        )



from pathlib import Path

import yaml
from yaml.parser import ParserError
from yamllint import linter
from yamllint.config import YamlLintConfig


class FileProviderException(Exception):
    pass


class YamlFileProviderException(FileProviderException):
    pass


class FileNotFoundException(FileProviderException):
    pass


class InvalidYamlException(YamlFileProviderException):
    def __init__(self, path: str):
        super().__init__(f"""{path} is not valid YAML.""")


class DuplicateKeysException(YamlFileProviderException):
    def __init__(self, duplicate_keys: str):
        super().__init__(f"""Duplicate keys found in your config file: {duplicate_keys}.""")


class YamlFileProvider:
    def load(path: str) -> dict:
        """
        Raises:
            FileNotFoundException: file is not there
            InvalidYamlException: file contains invalid yaml
            DuplicateKeysException: yaml contains duplicate keys
        """
        if not Path(path).exists():
            raise FileNotFoundException(f"`{path}` is missing.")
        try:
            yaml_content = yaml.safe_load(Path(path).read_text())
        except ParserError:
            raise InvalidYamlException(path)

        if not yaml_content:
            return {}

        YamlFileProvider.lint_yaml_for_duplicate_keys(path)

        return yaml_content

    def write(path: str, contents: dict, comment: str = ""):
        with open(path, "w") as file:
            file.write(comment)
            yaml.dump(contents, file)

    @staticmethod
    def lint_yaml_for_duplicate_keys(path):
        duplicate_keys = []
        with open(path, "r") as yaml_file:
            file_contents = yaml_file.read()
            results = linter.run(
                file_contents, YamlLintConfig(yaml.dump({"rules": {"key-duplicates": "enable"}}))
            )
            duplicate_keys = [
                "\t"
                + f"Line {result.line}: {result.message}".replace(
                    " in mapping (key-duplicates)", ""
                )
                for result in results
            ]
        if duplicate_keys:
            raise DuplicateKeysException(",".join(duplicate_keys))



from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.terraform_manifest import TerraformManifestProvider
from dbt_platform_helper.utils.versioning import (
    get_required_terraform_platform_modules_version,
)


class TerraformEnvironmentException(PlatformException):
    pass


class EnvironmentNotFoundException(TerraformEnvironmentException):
    pass


class TerraformEnvironment:
    def __init__(
        self,
        config_provider,
        manifest_provider: TerraformManifestProvider = None,
        io: ClickIOProvider = ClickIOProvider(),
    ):
        self.io = io
        self.config_provider = config_provider
        self.manifest_provider = manifest_provider or TerraformManifestProvider()

    def generate(self, environment_name, terraform_platform_modules_version_override=None):
        config = self.config_provider.get_enriched_config()

        if environment_name not in config.get("environments").keys():
            raise EnvironmentNotFoundException(
                f"cannot generate terraform for environment {environment_name}.  It does not exist in your configuration"
            )

        platform_config_terraform_modules_default_version = config.get("default_versions", {}).get(
            "terraform-platform-modules", ""
        )
        terraform_platform_modules_version = get_required_terraform_platform_modules_version(
            terraform_platform_modules_version_override,
            platform_config_terraform_modules_default_version,
        )

        self.manifest_provider.generate_environment_config(
            config, environment_name, terraform_platform_modules_version
        )



import itertools
import random
import re
import string
import traceback
from pathlib import Path
from typing import Callable
from typing import List
from typing import Union

import boto3
import click

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.load_balancers import ListenerNotFoundException
from dbt_platform_helper.providers.load_balancers import ListenerRuleNotFoundException
from dbt_platform_helper.providers.load_balancers import LoadBalancerNotFoundException
from dbt_platform_helper.providers.load_balancers import (
    get_https_listener_for_application,
)
from dbt_platform_helper.utils.application import Application
from dbt_platform_helper.utils.application import Environment
from dbt_platform_helper.utils.application import Service


class MaintenancePageException(PlatformException):
    pass


class LoadBalancedWebServiceNotFoundException(MaintenancePageException):
    def __init__(self, application_name: str):
        super().__init__(f"No services deployed yet to {application_name} ")


class FailedToActivateMaintenancePageException(MaintenancePageException):
    def __init__(
        self,
        application_name: str,
        env: str,
        original_exception: Exception,
        rolled_back_rules: dict[str, bool] = {},
    ):
        super().__init__(
            f"Maintenance page failed to activate for the {application_name} application in environment {env}."
        )
        self.orginal_exception = original_exception
        self.rolled_back_rules = rolled_back_rules

    def __str__(self):
        return (
            f"{super().__str__()}\n"
            f"Rolled-back rules: {self.rolled_back_rules }\n"
            f"Original exception: {self.orginal_exception}"
        )


def get_maintenance_page_type(session: boto3.Session, listener_arn: str) -> Union[str, None]:
    lb_client = session.client("elbv2")

    rules = lb_client.describe_rules(ListenerArn=listener_arn)["Rules"]
    tag_descriptions = get_rules_tag_descriptions(rules, lb_client)

    maintenance_page_type = None
    for description in tag_descriptions:
        tags = {t["Key"]: t["Value"] for t in description["Tags"]}
        if tags.get("name") == "MaintenancePage":
            maintenance_page_type = tags.get("type")

    return maintenance_page_type


def get_env_ips(vpc: str, application_environment: Environment) -> List[str]:
    account_name = f"{application_environment.session.profile_name}-vpc"
    vpc_name = vpc if vpc else account_name
    ssm_client = application_environment.session.client("ssm")

    try:
        param_value = ssm_client.get_parameter(Name=f"/{vpc_name}/EGRESS_IPS")["Parameter"]["Value"]
    except ssm_client.exceptions.ParameterNotFound:
        click.secho(f"No parameter found with name: /{vpc_name}/EGRESS_IPS")
        raise click.Abort

    return [ip.strip() for ip in param_value.split(",")]


def add_maintenance_page(
    session: boto3.Session,
    listener_arn: str,
    app: str,
    env: str,
    services: List[Service],
    allowed_ips: List[str],
    template: str = "default",
):
    lb_client = session.client("elbv2")
    maintenance_page_content = get_maintenance_page_template(template)
    bypass_value = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))

    rule_priority = itertools.count(start=1)
    maintenance_page_host_header_conditions = []
    try:
        for svc in services:
            target_group_arn = find_target_group(app, env, svc.name, session)

            # not all of an application's services are guaranteed to have been deployed to an environment
            if not target_group_arn:
                continue

            service_conditions = get_host_header_conditions(
                lb_client, listener_arn, target_group_arn
            )

            for ip in allowed_ips:
                create_header_rule(
                    lb_client,
                    listener_arn,
                    target_group_arn,
                    "X-Forwarded-For",
                    [ip],
                    "AllowedIps",
                    next(rule_priority),
                    service_conditions,
                )
                create_source_ip_rule(
                    lb_client,
                    listener_arn,
                    target_group_arn,
                    [ip],
                    "AllowedSourceIps",
                    next(rule_priority),
                    service_conditions,
                )

            create_header_rule(
                lb_client,
                listener_arn,
                target_group_arn,
                "Bypass-Key",
                [bypass_value],
                "BypassIpFilter",
                next(rule_priority),
                service_conditions,
            )

            # add to accumilating list of conditions for maintenace page rule
            maintenance_page_host_header_conditions.extend(service_conditions)

        click.secho(
            f"\nUse a browser plugin to add `Bypass-Key` header with value {bypass_value} to your requests. For more detail, visit https://platform.readme.trade.gov.uk/next-steps/put-a-service-under-maintenance/",
        )

        lb_client.create_rule(
            ListenerArn=listener_arn,
            Priority=next(rule_priority),
            Conditions=[
                {
                    "Field": "path-pattern",
                    "PathPatternConfig": {"Values": ["/*"]},
                },
                {
                    "Field": "host-header",
                    "HostHeaderConfig": {
                        "Values": sorted(
                            list(
                                {
                                    value
                                    for condition in maintenance_page_host_header_conditions
                                    for value in condition["HostHeaderConfig"]["Values"]
                                }
                            )
                        )
                    },
                },
            ],
            Actions=[
                {
                    "Type": "fixed-response",
                    "FixedResponseConfig": {
                        "StatusCode": "503",
                        "ContentType": "text/html",
                        "MessageBody": maintenance_page_content,
                    },
                }
            ],
            Tags=[
                {"Key": "name", "Value": "MaintenancePage"},
                {"Key": "type", "Value": template},
            ],
        )
    except Exception as e:
        deleted_rules = clean_up_maintenance_page_rules(session, listener_arn)
        raise FailedToActivateMaintenancePageException(
            app, env, f"{e}:\n {traceback.format_exc()}", deleted_rules
        )


def clean_up_maintenance_page_rules(
    session: boto3.Session, listener_arn: str, fail_when_not_deleted: bool = False
) -> dict[str, bool]:
    lb_client = session.client("elbv2")

    rules = lb_client.describe_rules(ListenerArn=listener_arn)["Rules"]
    tag_descriptions = get_rules_tag_descriptions(rules, lb_client)

    deletes = {}
    for name in ["MaintenancePage", "AllowedIps", "BypassIpFilter", "AllowedSourceIps"]:
        deleted = delete_listener_rule(tag_descriptions, name, lb_client)
        deletes[name] = bool(deleted)
        if fail_when_not_deleted and name == "MaintenancePage" and not deleted:
            raise ListenerRuleNotFoundException()

    return deletes


def remove_maintenance_page(session: boto3.Session, listener_arn: str):
    clean_up_maintenance_page_rules(session, listener_arn, True)


class MaintenancePage:
    def __init__(
        self,
        application: Application,
        io: ClickIOProvider = ClickIOProvider(),
        get_https_listener_for_application: Callable[
            [boto3.Session, str, str], str
        ] = get_https_listener_for_application,
        # TODO refactor get_maintenance_page_type, add_maintenance_page, remove_maintenance_page into MaintenancePage class with LoadBalancerProvider as the dependency
        get_maintenance_page_type: Callable[
            [boto3.Session, str], Union[str, None]
        ] = get_maintenance_page_type,
        get_env_ips: Callable[[str, Environment], List[str]] = get_env_ips,
        add_maintenance_page: Callable[
            [boto3.Session, str, str, str, List[Service], tuple, str], None
        ] = add_maintenance_page,
        remove_maintenance_page: Callable[[boto3.Session, str], None] = remove_maintenance_page,
    ):
        self.application = application
        self.get_https_listener_for_application = get_https_listener_for_application
        self.io = io
        self.get_maintenance_page_type = get_maintenance_page_type
        self.get_env_ips = get_env_ips
        self.add_maintenance_page = add_maintenance_page
        self.remove_maintenance_page = remove_maintenance_page

    def _get_deployed_load_balanced_web_services(self, app: Application, svc: List[str]):
        if "*" in svc:
            services = [s for s in app.services.values() if s.kind == "Load Balanced Web Service"]
        else:
            all_services = [get_app_service(app, s) for s in list(svc)]
            services = [s for s in all_services if s.kind == "Load Balanced Web Service"]
        if not services:
            raise LoadBalancedWebServiceNotFoundException(app.name)
        return services

    def activate(self, env: str, services: List[str], template: str, vpc: Union[str, None]):

        services = self._get_deployed_load_balanced_web_services(self.application, services)
        application_environment = get_app_environment(self.application, env)

        try:
            https_listener = self.get_https_listener_for_application(
                application_environment.session, self.application.name, env
            )
            current_maintenance_page = self.get_maintenance_page_type(
                application_environment.session, https_listener
            )
            remove_current_maintenance_page = False
            if current_maintenance_page:
                remove_current_maintenance_page = self.io.confirm(
                    f"There is currently a '{current_maintenance_page}' maintenance page for the {env} "
                    f"environment in {self.application.name}.\nWould you like to replace it with a '{template}' "
                    f"maintenance page?"
                )
                if not remove_current_maintenance_page:
                    return

            if remove_current_maintenance_page or self.io.confirm(
                f"You are about to enable the '{template}' maintenance page for the {env} "
                f"environment in {self.application.name}.\nWould you like to continue?"
            ):
                if current_maintenance_page and remove_current_maintenance_page:
                    self.remove_maintenance_page(application_environment.session, https_listener)

                allowed_ips = self.get_env_ips(vpc, application_environment)

                self.add_maintenance_page(
                    application_environment.session,
                    https_listener,
                    self.application.name,
                    env,
                    services,
                    allowed_ips,
                    template,
                )
                self.io.info(
                    f"Maintenance page '{template}' added for environment {env} in application {self.application.name}",
                )

        except LoadBalancerNotFoundException:
            # TODO push exception to command layer
            self.io.abort_with_error(
                f"No load balancer found for environment {env} in the application {self.application.name}.",
            )

        except ListenerNotFoundException:
            # TODO push exception to command layer
            self.io.abort_with_error(
                f"No HTTPS listener found for environment {env} in the application {self.application.name}.",
            )

    def deactivate(self, env: str):
        application_environment = get_app_environment(self.application, env)

        try:
            https_listener = self.get_https_listener_for_application(
                application_environment.session, self.application.name, env
            )
            current_maintenance_page = self.get_maintenance_page_type(
                application_environment.session, https_listener
            )

            # TODO discuss, reduce number of return statements but more nested if statements
            if not current_maintenance_page:
                self.io.warn("There is no current maintenance page to remove")
                return

            if not self.io.confirm(
                f"There is currently a '{current_maintenance_page}' maintenance page, "
                f"would you like to remove it?"
            ):
                return

            self.remove_maintenance_page(application_environment.session, https_listener)
            self.io.info(
                f"Maintenance page removed from environment {env} in application {self.application.name}",
            )

        except LoadBalancerNotFoundException:
            # TODO push exception to command layer
            self.io.abort_with_error(
                f"No load balancer found for environment {env} in the application {self.application.name}.",
            )

        except ListenerNotFoundException:
            # TODO push exception to command layer
            self.io.abort_with_error(
                f"No HTTPS listener found for environment {env} in the application {self.application.name}.",
            )


def get_app_service(application: Application, svc_name: str) -> Service:
    application_service = application.services.get(svc_name)

    if not application_service:
        # TODO raise exception instead of abort
        click.secho(
            f"The service {svc_name} was not found in the application {application.name}. "
            f"It either does not exist, or has not been deployed.",
            fg="red",
        )
        raise click.Abort

    return application_service


def get_app_environment(application: Application, env_name: str) -> Environment:
    application_environment = application.environments.get(env_name)

    if not application_environment:
        click.secho(
            f"The environment {env_name} was not found in the application {application.name}. "
            f"It either does not exist, or has not been deployed.",
            fg="red",
        )
        raise click.Abort

    return application_environment


def get_rules_tag_descriptions(rules: list, lb_client):
    tag_descriptions = []
    chunk_size = 20

    for i in range(0, len(rules), chunk_size):
        chunk = rules[i : i + chunk_size]
        resource_arns = [r["RuleArn"] for r in chunk]
        response = lb_client.describe_tags(ResourceArns=resource_arns)
        tag_descriptions.extend(response["TagDescriptions"])

    return tag_descriptions


def delete_listener_rule(tag_descriptions: list, tag_name: str, lb_client: boto3.client):
    current_rule_arn = None

    for description in tag_descriptions:
        tags = {t["Key"]: t["Value"] for t in description["Tags"]}
        if tags.get("name") == tag_name:
            current_rule_arn = description["ResourceArn"]
            if current_rule_arn:
                lb_client.delete_rule(RuleArn=current_rule_arn)

    return current_rule_arn


def get_maintenance_page_template(template) -> str:
    template_contents = (
        Path(__file__)
        .parent.parent.joinpath(
            f"templates/svc/maintenance_pages/{template}.html",
        )
        .read_text()
        .replace("\n", "")
    )

    # [^\S]\s+ - Remove any space that is not preceded by a non-space character.
    return re.sub(r"[^\S]\s+", "", template_contents)


def find_target_group(app: str, env: str, svc: str, session: boto3.Session) -> str:
    rg_tagging_client = session.client("resourcegroupstaggingapi")
    response = rg_tagging_client.get_resources(
        TagFilters=[
            {
                "Key": "copilot-application",
                "Values": [
                    app,
                ],
                "Key": "copilot-environment",
                "Values": [
                    env,
                ],
                "Key": "copilot-service",
                "Values": [
                    svc,
                ],
            },
        ],
        ResourceTypeFilters=[
            "elasticloadbalancing:targetgroup",
        ],
    )
    for resource in response["ResourceTagMappingList"]:
        tags = {tag["Key"]: tag["Value"] for tag in resource["Tags"]}

        if (
            "copilot-service" in tags
            and tags["copilot-service"] == svc
            and "copilot-environment" in tags
            and tags["copilot-environment"] == env
            and "copilot-application" in tags
            and tags["copilot-application"] == app
        ):
            return resource["ResourceARN"]

    click.secho(
        f"No target group found for application: {app}, environment: {env}, service: {svc}",
        fg="red",
    )

    return None


def create_header_rule(
    lb_client: boto3.client,
    listener_arn: str,
    target_group_arn: str,
    header_name: str,
    values: list,
    rule_name: str,
    priority: int,
    conditions: list,
):
    # add new condition to existing conditions
    combined_conditions = [
        {
            "Field": "http-header",
            "HttpHeaderConfig": {"HttpHeaderName": header_name, "Values": values},
        }
    ] + conditions

    lb_client.create_rule(
        ListenerArn=listener_arn,
        Priority=priority,
        Conditions=combined_conditions,
        Actions=[{"Type": "forward", "TargetGroupArn": target_group_arn}],
        Tags=[
            {"Key": "name", "Value": rule_name},
        ],
    )

    click.secho(
        f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\n\nIf request header {header_name} contains one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.",
        fg="green",
    )


def normalise_to_cidr(ip: str):
    if "/" in ip:
        return ip
    SINGLE_IPV4_CIDR_PREFIX_LENGTH = "32"
    return f"{ip}/{SINGLE_IPV4_CIDR_PREFIX_LENGTH}"


def create_source_ip_rule(
    lb_client: boto3.client,
    listener_arn: str,
    target_group_arn: str,
    values: list,
    rule_name: str,
    priority: int,
    conditions: list,
):
    # add new condition to existing conditions

    combined_conditions = [
        {
            "Field": "source-ip",
            "SourceIpConfig": {"Values": [normalise_to_cidr(value) for value in values]},
        }
    ] + conditions

    lb_client.create_rule(
        ListenerArn=listener_arn,
        Priority=priority,
        Conditions=combined_conditions,
        Actions=[{"Type": "forward", "TargetGroupArn": target_group_arn}],
        Tags=[
            {"Key": "name", "Value": rule_name},
        ],
    )

    click.secho(
        f"Creating listener rule {rule_name} for HTTPS Listener with arn {listener_arn}.\n\nIf request source ip matches one of the values {values}, the request will be forwarded to target group with arn {target_group_arn}.",
        fg="green",
    )


def get_host_header_conditions(
    lb_client: boto3.client, listener_arn: str, target_group_arn: str
) -> list:
    rules = lb_client.describe_rules(ListenerArn=listener_arn)["Rules"]

    # Get current set of forwarding conditions for the target group
    for rule in rules:
        for action in rule["Actions"]:
            if action["Type"] == "forward" and action["TargetGroupArn"] == target_group_arn:
                conditions = rule["Conditions"]

    # filter to host-header conditions
    conditions = [
        {i: condition[i] for i in condition if i != "Values"}
        for condition in conditions
        if condition["Field"] == "host-header"
    ]

    # remove internal hosts
    conditions[0]["HostHeaderConfig"]["Values"] = [
        v for v in conditions[0]["HostHeaderConfig"]["Values"]
    ]

    return conditions



from copy import deepcopy
from pathlib import Path

from schema import SchemaError

from dbt_platform_helper.constants import PLATFORM_CONFIG_FILE
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.config_validator import ConfigValidatorError
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.platform_config_schema import PlatformConfigSchema
from dbt_platform_helper.providers.yaml_file import FileNotFoundException
from dbt_platform_helper.providers.yaml_file import FileProviderException
from dbt_platform_helper.providers.yaml_file import YamlFileProvider


class ConfigProvider:
    def __init__(
        self,
        config_validator: ConfigValidator = None,
        file_provider: YamlFileProvider = None,
        io: ClickIOProvider = None,
    ):
        self.config = {}
        self.validator = config_validator or ConfigValidator()
        self.io = io or ClickIOProvider()
        self.file_provider = file_provider or YamlFileProvider

    # TODO refactor so that apply_environment_defaults isn't set, discarded and set again
    def get_enriched_config(self):
        return self.apply_environment_defaults(self.load_and_validate_platform_config())

    def _validate_platform_config(self):
        PlatformConfigSchema.schema().validate(self.config)

        # TODO= logically this isn't validation but loading + parsing, to move.
        # also, we apply defaults but discard that data.  Should we just apply
        # defaults to config returned by load_and_validate
        enriched_config = ConfigProvider.apply_environment_defaults(self.config)
        try:
            self.validator.run_validations(enriched_config)
        except ConfigValidatorError as exc:
            self.io.abort_with_error(f"Config validation has failed.\n{str(exc)}")

    def load_and_validate_platform_config(self, path=PLATFORM_CONFIG_FILE):
        try:
            self.config = self.file_provider.load(path)
        except FileNotFoundException as e:
            self.io.abort_with_error(
                f"{e} Please check it exists and you are in the root directory of your deployment project."
            )
        except FileProviderException as e:
            self.io.abort_with_error(f"Error loading configuration from {path}: {e}")

        try:
            self._validate_platform_config()
        except SchemaError as e:
            self.io.abort_with_error(f"Schema error in {path}. {e}")

        return self.config

    def load_unvalidated_config_file(self, path=PLATFORM_CONFIG_FILE):
        try:
            return self.file_provider.load(path)
        except FileProviderException:
            return {}

    # TODO this general function should be moved out of ConfigProvider
    def config_file_check(self, path=PLATFORM_CONFIG_FILE):
        if not Path(path).exists():
            self.io.abort_with_error(
                f"`{path}` is missing. "
                "Please check it exists and you are in the root directory of your deployment project."
            )

    @staticmethod
    def apply_environment_defaults(config):
        if "environments" not in config:
            return config

        enriched_config = deepcopy(config)

        environments = enriched_config["environments"]
        env_defaults = environments.get("*", {})
        without_defaults_entry = {
            name: data if data else {} for name, data in environments.items() if name != "*"
        }

        default_versions = config.get("default_versions", {})

        def combine_env_data(data):
            return {
                **env_defaults,
                **data,
                "versions": {
                    **default_versions,
                    **env_defaults.get("versions", {}),
                    **data.get("versions", {}),
                },
            }

        defaulted_envs = {
            env_name: combine_env_data(env_data)
            for env_name, env_data in without_defaults_entry.items()
        }

        enriched_config["environments"] = defaulted_envs

        return enriched_config



import os
from datetime import datetime

from dbt_platform_helper.providers.yaml_file import YamlFileProvider


class CacheProvider:
    def __init__(
        self,
        file_provider: YamlFileProvider = None,
    ):
        self._cache_file = ".platform-helper-config-cache.yml"
        self.file_provider = file_provider or YamlFileProvider

    def read_supported_versions_from_cache(self, resource_name):

        platform_helper_config = self.file_provider.load(self._cache_file)

        return platform_helper_config.get(resource_name).get("versions")

    def update_cache(self, resource_name, supported_versions):

        platform_helper_config = {}

        if self.__cache_exists():
            platform_helper_config = self.file_provider.load(self._cache_file)

        cache_dict = {
            resource_name: {
                "versions": supported_versions,
                "date-retrieved": datetime.now().strftime("%d-%m-%y %H:%M:%S"),
            }
        }

        platform_helper_config.update(cache_dict)

        self.file_provider.write(
            self._cache_file,
            platform_helper_config,
            "# [!] This file is autogenerated via the platform-helper. Do not edit.\n",
        )

    def cache_refresh_required(self, resource_name) -> bool:
        """
        Checks if the platform-helper should reach out to AWS to 'refresh' its
        cached values.

        An API call is needed if any of the following conditions are met:
            1. No cache file (.platform-helper-config.yml) exists.
            2. The resource name (e.g. redis, opensearch) does not exist within the cache file.
            3. The date-retrieved value of the cached data is > than a time interval. In this case 1 day.
        """

        if not self.__cache_exists():
            return True

        platform_helper_config = self.file_provider.load(self._cache_file)

        if platform_helper_config.get(resource_name):
            return self.__check_if_cached_datetime_is_greater_than_interval(
                platform_helper_config[resource_name].get("date-retrieved"), 1
            )

        return True

    @staticmethod
    def __check_if_cached_datetime_is_greater_than_interval(date_retrieved, interval_in_days):

        current_datetime = datetime.now()
        cached_datetime = datetime.strptime(date_retrieved, "%d-%m-%y %H:%M:%S")
        delta = current_datetime - cached_datetime

        return delta.days > interval_in_days

    def __cache_exists(self):
        return os.path.exists(self._cache_file)



import click

from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.versioning import RequiredVersion


@click.group(chain=True, cls=ClickDocOptGroup)
def version():
    """Contains subcommands for getting version information about the current
    project."""


@version.command(help="Print the version of platform-tools required by the current project")
@click.option(
    "--pipeline",
    required=False,
    type=str,
    help="Take into account platform-tools version overrides in the specified pipeline",
)
def get_platform_helper_for_project(pipeline):
    """
    Version precedence is in this order:
        - if the --pipeline option is supplied, the version in 'platform-config.yml' in:
            environment_pipelines:
                <pipeline>:
                    ...
                    versions:
                        platform-helper
        - The version from default_versions/platform-helper in 'platform-config.yml'
        - Fall back on the version in the deprecated '.platform-helper-version' file
    """
    try:
        RequiredVersion().get_required_version(pipeline)
    except PlatformException as err:
        ClickIOProvider().abort_with_error(str(err))



import click
import yaml
from jinja2 import Environment
from jinja2 import FileSystemLoader

from dbt_platform_helper.providers.files import FileProvider


def to_yaml(value):
    return yaml.dump(value, sort_keys=False)


# TODO - extract file provider functionality from this - and figure out what it actually does!
# Move to the new file provider - or potentially copilot?
def generate_override_files(base_path, file_path, output_dir):
    def generate_files_for_dir(pattern):
        for file in file_path.glob(pattern):
            if file.is_file():
                contents = file.read_text()
                file_name = str(file).removeprefix(f"{file_path}/")
                click.echo(
                    FileProvider.mkfile(
                        base_path,
                        output_dir / file_name,
                        contents,
                        overwrite=True,
                    )
                )

    generate_files_for_dir("*")
    generate_files_for_dir("bin/*")


def generate_override_files_from_template(base_path, overrides_path, output_dir, template_data={}):
    templates = Environment(
        loader=FileSystemLoader(f"{overrides_path}"), keep_trailing_newline=True
    )
    environments = ",".join([env["name"] for env in template_data["environments"]])
    data = {"environments": environments}

    def generate_files_for_dir(pattern):

        for file in overrides_path.glob(pattern):
            if file.is_file():
                file_name = str(file).removeprefix(f"{overrides_path}/")
                contents = templates.get_template(str(file_name)).render(data)
                message = FileProvider.mkfile(
                    base_path, output_dir / file_name, contents, overwrite=True
                )
                click.echo(message)

    generate_files_for_dir("*")
    generate_files_for_dir("bin/*")



import click

from dbt_platform_helper.platform_exception import PlatformException


class ClickIOProvider:
    def warn(self, message: str):
        click.secho(message, fg="magenta")

    def error(self, message: str):
        click.secho(f"Error: {message}", fg="red")

    def info(self, message: str):
        click.secho(message)

    def input(self, message: str) -> str:
        return click.prompt(message)

    def confirm(self, message: str) -> bool:
        try:
            return click.confirm(message)
        except click.Abort:
            raise ClickIOProviderException(message + " [y/N]: Error: invalid input")

    def abort_with_error(self, message: str):
        click.secho(f"Error: {message}", err=True, fg="red")
        exit(1)


class ClickIOProviderException(PlatformException):
    pass



from dbt_platform_helper.providers.cache import CacheProvider


class RedisProvider:
    def __init__(self, elasticache_client):
        self.elasticache_client = elasticache_client

    def get_supported_redis_versions(self):

        cache_provider = self.__get_cache_provider()

        if cache_provider.cache_refresh_required("redis"):

            supported_versions_response = self.elasticache_client.describe_cache_engine_versions(
                Engine="redis"
            )

            supported_versions = [
                version["EngineVersion"]
                for version in supported_versions_response["CacheEngineVersions"]
            ]

            cache_provider.update_cache("redis", supported_versions)

            return supported_versions

        else:
            return cache_provider.read_supported_versions_from_cache("redis")

    # TODO - cache provider instantiated here rather than via dependancy injection since it will likely only be used in the get_supported_redis_versions method.
    # If another method is added which needs a CacheProvider, it should be injected into the constructor instead.
    @staticmethod
    def __get_cache_provider():
        return CacheProvider()



import click
from slack_sdk import WebClient
from slack_sdk.models import blocks

from dbt_platform_helper.utils.arn_parser import ARN
from dbt_platform_helper.utils.click import ClickDocOptGroup
from dbt_platform_helper.utils.versioning import (
    check_platform_helper_version_needs_update,
)


@click.group(cls=ClickDocOptGroup, help="Send Slack notifications")
def notify():
    check_platform_helper_version_needs_update()


@notify.command(
    help="Send environment progress notifications. This creates (or updates if --slack-ref is provided) the top level message to the channel."
)
@click.argument("slack-channel-id")
@click.argument("slack-token")
@click.argument("message")
@click.option("--build-arn")
@click.option("--repository")
@click.option("--commit-sha")
@click.option("--slack-ref", help="Slack message reference of the message to update")
def environment_progress(
    slack_channel_id: str,
    slack_token: str,
    message: str,
    build_arn: str,
    repository: str,
    commit_sha: str,
    slack_ref: str,
):
    args = _get_slack_args(build_arn, commit_sha, message, repository, slack_channel_id)
    slack = _get_slack_client(slack_token)

    if slack_ref:
        response = slack.chat_update(ts=slack_ref, **args)
    else:
        response = slack.chat_postMessage(ts=slack_ref, **args)

    print(response["ts"])


def _get_slack_args(
    build_arn: str, commit_sha: str, message: str, repository: str, slack_channel_id: str
):
    context_elements = []
    if repository:
        context_elements.append(f"*Repository*: <https://github.com/{repository}|{repository}>")
        if commit_sha:
            context_elements.append(
                f"*Revision*: <https://github.com/{repository}/commit/{commit_sha}|{commit_sha}>"
            )
    if build_arn:
        context_elements.append(f"<{get_build_url(build_arn)}|Build Logs>")
    message_blocks = [
        blocks.SectionBlock(
            text=blocks.TextObject(type="mrkdwn", text=message),
        ),
    ]

    if context_elements:
        message_blocks.append(
            blocks.ContextBlock(
                elements=[
                    blocks.TextObject(type="mrkdwn", text=element) for element in context_elements
                ]
            )
        )

    args = {
        "channel": slack_channel_id,
        "blocks": message_blocks,
        "text": message,
        "unfurl_links": False,
        "unfurl_media": False,
    }
    return args


def _get_slack_client(token: str):
    return WebClient(token=token)


@notify.command(help="Add a comment to an existing Slack message")
@click.argument("slack-channel-id")
@click.argument("slack-token")
@click.argument("slack-ref")
@click.argument("message")
@click.option("--title", default=None, help="Message title")
@click.option("--send-to-main-channel", default=False, help="Send to main channel")
def add_comment(
    slack_channel_id: str,
    slack_token: str,
    slack_ref: str,
    message: str,
    title: str,
    send_to_main_channel: bool,
):
    slack = _get_slack_client(slack_token)

    slack.chat_postMessage(
        channel=slack_channel_id,
        blocks=[blocks.SectionBlock(text=blocks.TextObject(type="mrkdwn", text=message))],
        text=title if title else message,
        reply_broadcast=send_to_main_channel,
        unfurl_links=False,
        unfurl_media=False,
        thread_ts=slack_ref,
    )


def get_build_url(build_arn: str):
    try:
        arn = ARN(build_arn)
        url = (
            "https://{region}.console.aws.amazon.com/codesuite/codebuild/{account}/projects/{"
            "project}/build/{project}%3A{build_id}"
        )
        return url.format(
            region=arn.region,
            account=arn.account_id,
            project=arn.project.replace("build/", ""),
            build_id=arn.build_id,
        )
    except ValueError:
        return ""
